#!/usr/bin/env python3
"""Zaza functional tests for prometheus."""
import json
import unittest

import requests

import yaml

from zaza import model
from zaza.charm_lifecycle import utils as lifecycle_utils


class PrometheusCharmTestBase(unittest.TestCase):
    """Base class for functional tests."""

    @classmethod
    def setUpClass(cls):
        """Set up test case."""
        cls.model_name = model.get_juju_model()
        cls.test_config = lifecycle_utils.get_charm_config()
        model.block_until_all_units_idle()
        cls.leader_addr = model.get_lead_unit_ip("prometheus2")

    def cat_unit(self, unit, path):
        """Run cat command on the unit."""
        unit_res = model.run_on_unit(unit, "sudo cat {}".format(path))
        return unit_res["Stdout"]


class BasicPrometheusCharmTest(PrometheusCharmTestBase):
    """Class for functional tests."""

    def test_service_running(self):
        """Test if we have prom running on all units."""
        running_for = model.get_unit_service_start_time(
            "prometheus2/0", "prometheus", timeout=20
        )
        self.assertGreater(running_for, 0)

    def test_global_prometheus_yml(self):
        """Test global prometheus.yml config values."""
        yml = yaml.safe_load(
            self.cat_unit(
                "prometheus2/0", "/var/snap/prometheus/current/prometheus.yml"
            )
        )
        self.assertEqual(yml["global"]["scrape_interval"], "30s")
        self.assertEqual(yml["global"]["scrape_timeout"], "20s")
        self.assertEqual(yml["global"]["evaluation_interval"], "15s")

    def test_prometheus_yml_scrape_configs(self):
        """Test scrape configurations."""
        scrape_configs = yaml.safe_load(
            self.cat_unit(
                "prometheus2/0", "/var/snap/prometheus/current/prometheus.yml"
            )
        )["scrape_configs"]
        expected_jobs = ["prometheus", "telegraf"]
        actual_jobs = [j["job_name"] for j in scrape_configs]
        self.assertEqual(sorted(actual_jobs), expected_jobs)
        telegraf_conf = [j for j in scrape_configs if j["job_name"] == "telegraf"][0][
            "static_configs"
        ]
        self.assertEqual(len(telegraf_conf), 1)
        labels = telegraf_conf[0]["labels"]
        self.assertTrue("dns_name" in labels)
        result = model.run_on_leader(
            "prometheus2", "dig +short {}".format(labels["dns_name"])
        )
        label_addr = result["Stdout"].strip()
        telegraf_addr = model.get_lead_unit_ip("telegraf")
        self.assertEqual(label_addr, telegraf_addr)

    def test_service_reload_after_change(self):
        """Test service reload."""
        model.add_unit("ubuntu", wait_appear=True)
        model.block_until_all_units_idle(timeout=600)

        action = model.run_on_unit("ubuntu/1", "hostname -f")
        new_target = action["Stdout"].strip()

        prom_ip = model.get_lead_unit_ip("prometheus2")
        prom_port = 9090
        url = "http://{}:{}/api/v1/targets".format(prom_ip, prom_port)
        response = requests.get(url)
        content = json.loads(response.content)

        target_found = False
        for targets in content["data"]["activeTargets"]:
            if targets["labels"].get("dns_name", None) == new_target:
                target_found = True

        self.assertTrue(target_found)
