"""Unit tests for prometheus2 reactive charm."""
import os
import shutil
import sys
import tempfile
import unittest

from charmhelpers.core import hookenv, host, unitdata
from charmhelpers.core.templating import render

from charms.reactive import bus

import mock

import yaml


mock_charms = mock.MagicMock()
sys.modules["charms.layer"] = mock_charms

from reactive import prometheus as react_prom  # NOQA E402


os.environ["JUJU_UNIT_NAME"] = "prometheus"
os.environ["CHARM_DIR"] = ".."

fixed_scrape_config = {
    "job_name": "prometheus",
    "metrics_path": "/metrics",
    "static_configs": [{"targets": ["localhost:9090"]}],
}


class ReactInterfaceMock(object):
    """Mock for reactive interface."""

    def __init__(self, serv_dict):
        """Set service dictionary."""
        self.serv_dict = serv_dict

    def services(self):
        """Return services list."""
        srvs = []
        for name, hostports in self.serv_dict.items():
            srvs.append(
                {
                    "service_name": name,
                    "hosts": [{"hostname": h, "port": p} for h, p in hostports],
                }
            )
        return srvs


class UnitdataKvMock(object):
    """Mock unitdata keyvalue db."""

    def __init__(self, kvdata):
        """Set kvdata."""
        self.kvdata = kvdata
        self.itercount = 0

    def set(self, *args, **kwargs):
        """Do nothing."""
        pass

    def get(self, key, *args, **kwargs):
        """Return value for key or empty list."""
        return self.kvdata.get(key, [])

    def getrange(self, *args, **kwargs):
        """Return empty dict."""
        return {}

    def unset(self, *args, **kwargs):
        """Do nothing."""
        pass


class SimpleConfigMock(dict):
    """Mock charm config."""

    def __init__(self, *arg, **kw):
        """Initialize instance params."""
        super(SimpleConfigMock, self).__init__(*arg, **kw)
        self._changed_dict = {}
        self._changed_default = True

    def changed(self, key):
        """Return true if a key has changed or is new."""
        return self._changed_dict.get(key, self._changed_default)

    def set_changed(self, changed_dict):
        """Update the changed dictionary."""
        self._changed_dict.update(changed_dict)


class TestPrometheusContext(unittest.TestCase):
    """Test Prometheus hooks."""

    def setUp(self):
        """Set up test case."""
        super(TestPrometheusContext, self).setUp()
        self._init_load_config_yml_defaults()
        self._init_override_render_for_nonroot()
        self._init_common_mocks()
        self._init_tempdir_and_filenames()

    def _init_common_mocks(self):
        """Initialize common mocks."""
        # There are too many common mocks, so rather than having
        # an insane number of class decorators and a growing number
        # of test arguments, do them all here. Tests can still access
        # the mocks via the mock_* attribute or in the mocked location
        # (so hookenv.unit_get.return_value='foo' is fine)
        def mockit(dotpath):
            p = mock.patch(dotpath)
            m = p.start()
            self.addCleanup(p.stop)
            return m

        mockit("os.chown")
        mockit("os.fchown")
        mockit("reactive.prometheus.hookenv.open_port")
        mockit("reactive.prometheus.hookenv.close_port")
        mockit("reactive.prometheus.host.service_restart")
        mockit("reactive.prometheus.host.service_start")
        mockit("reactive.prometheus.host.service_running")
        mockit("reactive.prometheus.data_changed")
        mockit("reactive.prometheus.validate_config")
        mockit("reactive.prometheus.hookenv.unit_get")
        mockit("reactive.prometheus.hookenv.config")
        mockit("charmhelpers.core.hookenv.log")
        mockit("charmhelpers.core.host.log")
        mockit("reactive.prometheus.get_version")
        mockit("reactive.prometheus.socket")

        assert self.def_config, "Called too early"
        hookenv.config.return_value = self.def_config
        hookenv.unit_get.return_value = "localhost"
        react_prom.get_version.return_value = (2, 0, 0)
        react_prom.socket.gethostbyaddr.return_value = ["a-dns-name.dom"]

    def _init_tempdir_and_filenames(self):
        self.dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.dir)
        self.maxDiff = None
        self.prom_yml = os.path.join(self.dir, "test_prometheus.yml")
        self.prom_def = os.path.join(self.dir, "test_daemon_arguments")
        self.prom_custom_rules = os.path.join(self.dir, "test_custom.rules")
        react_prom.PATHS = {
            "prometheus_yml_tmpl": "prometheus.yml.j2",
            "prometheus_def_tmpl": "etc_default_prometheus.j2",
            "prometheus_rules_cron": os.path.join(
                self.dir, "test_prometheus-rules-refresh"
            ),
            "prometheus_ssh_keyfile": os.path.join(self.dir, "test_prometheus_git_key"),
            "refresh_rules_script": os.path.join(self.dir, "test_refresh_rules.sh"),
            "refresh_rules_script_tmpl": "refresh_rules.sh.j2",
            "prometheus_yml": self.prom_yml,
            "prometheus_def": self.prom_def,
            "prometheus_promreg_tgts_yml": os.path.join(self.dir, "test_targets.yaml"),
            "custom_rules_path": self.prom_custom_rules,
            "juju_rules_path": os.path.join(self.dir, "test_juju.rules"),
            "generic_rules_path": os.path.join(self.dir, "test_generic.rules"),
            "generic_rules_tmpl": "generic.rules.j2",
            "external_rules_path": os.path.join(self.dir, "test_custom.rules.d"),
            "promtool": "/snap/bin/prometheus.promtool",
            "service_name": "snap.prometheus.prometheus",
            "promreg_yml": os.path.join(self.dir, "test_promreg.yaml"),
            "promreg_tgts_yml": os.path.join(self.dir, "test_targets.yaml"),
            "promreg_service_name": "snap.promreg.promreg",
            "certs_dir": os.path.join(self.dir, "test_certs"),
        }
        # ugly hack, to avoid carrying global unitdata state across tests
        os.environ["UNIT_STATE_DB"] = os.path.join(self.dir, ".unit-state.db")
        unitdata._KV = None

    def _init_load_config_yml_defaults(self):
        # create def_config with parsed config.yaml defaults
        with open("config.yaml") as fh:
            conf_yaml = yaml.safe_load(fh)
            self.def_config = SimpleConfigMock(
                {k: v["default"] for k, v in conf_yaml["options"].items()}
            )

    def _init_override_render_for_nonroot(self):
        # need to override render() for 644 perms (ie allow overriding target
        # when running as non-root)
        def render_and_chmod(*args, **kwargs):
            kwargs["templates_dir"] = os.path.join(
                hookenv.charm_dir(), "src", "templates"
            )
            render(*args, **kwargs)
            os.chmod(kwargs["target"], 0o664)

        self.mock_render = mock.patch(
            "reactive.prometheus.render", side_effect=render_and_chmod
        )
        self.mock_render.start()
        self.addCleanup(self.mock_render.stop)

    @mock.patch("reactive.prometheus.set_state")
    @mock.patch("reactive.prometheus.get_related_unit_names")
    @mock.patch("charmhelpers.core.unitdata.kv")
    def test_update_prometheus_targets(
        self, mock_kv, mock_related_units, mock_set_state
    ):
        """Verify prometheus targets update."""
        srv1_name = "foo"
        srv1_hostport1 = ["foohost1", "fooport1"]
        srv1_hostport2 = ["foohost2", "fooport2"]
        srv2_name = "bar"
        srv2_hostport1 = ["barhost1", "barport1"]
        srvs_mock = {
            srv1_name: [srv1_hostport1, srv1_hostport2],
            srv2_name: [srv2_hostport1],
        }
        mock_related_units.return_value = {
            "{0}:{1}".format(*srv1_hostport1): srv1_name,
            "{0}:{1}".format(*srv1_hostport2): srv1_name,
            "{0}:{1}".format(*srv2_hostport1): srv2_name,
        }
        req_mock = ReactInterfaceMock(srvs_mock)
        kv_data = {
            "target_jobs": [
                {
                    "job_name": srv1_name,
                    "targets": [
                        {
                            "hostname": "{0}:{1}".format(*srv1_hostport1),
                            "unit_name": "{}/0".format(srv1_name),
                            "dns_name": "a-dns-name.dom",
                        },
                        {
                            "hostname": "{0}:{1}".format(*srv1_hostport2),
                            "unit_name": "{}/1".format(srv1_name),
                            "dns_name": "a-dns-name.dom",
                        },
                    ],
                },
                {
                    "job_name": srv2_name,
                    "targets": [
                        {
                            "hostname": "{0}:{1}".format(*srv2_hostport1),
                            "unit_name": "{}/0".format(srv2_name),
                            "dns_name": "a-dns-name.dom",
                        }
                    ],
                },
            ],
            "scrape_jobs": [],
        }
        mock_kv.return_value = UnitdataKvMock(kv_data)
        react_prom.update_prometheus_targets(req_mock)  # Updates unitdata.kv
        mock_set_state.assert_called_once_with("prometheus.do-check-reconfig")
        # must read the kv in order to write proper yaml
        react_prom.write_prometheus_config_yml()
        react_prom.validate_config.assert_called_once_with()
        yaml_content = yaml.safe_load(open(self.prom_yml))
        exp_list = [
            fixed_scrape_config,
            {
                "job_name": srv1_name,
                "static_configs": [
                    {
                        "labels": {
                            "group": "promoagents-juju",
                            "dns_name": "a-dns-name.dom",
                        },
                        "targets": ["{0}:{1}".format(*srv1_hostport1)],
                    },
                    {
                        "targets": ["{0}:{1}".format(*srv1_hostport2)],
                        "labels": {
                            "group": "promoagents-juju",
                            "dns_name": "a-dns-name.dom",
                        },
                    },
                ],
            },
            {
                "job_name": srv2_name,
                "static_configs": [
                    {
                        "labels": {
                            "group": "promoagents-juju",
                            "dns_name": "a-dns-name.dom",
                        },
                        "targets": ["{0}:{1}".format(*srv2_hostport1)],
                    },
                ],
            },
        ]
        self.assertListEqual(
            sorted(yaml_content["scrape_configs"], key=lambda x: x["job_name"]),
            sorted(exp_list, key=lambda x: x["job_name"]),
        )

        return

    @mock.patch("reactive.prometheus.set_state")
    def test_update_prometheus_no_targets(self, mock_set_state):
        """Verify updating prometheus with no targets."""
        react_prom.update_prometheus_no_targets()
        mock_set_state.assert_called_once_with("prometheus.do-check-reconfig")
        react_prom.write_prometheus_config_yml()
        react_prom.write_prometheus_config_def()
        react_prom.validate_config.assert_called_once_with()
        yaml_content = yaml.safe_load(open(self.prom_yml))
        exp_dict = fixed_scrape_config
        self.assertDictEqual(yaml_content["scrape_configs"][0], exp_dict)

    def test_update_prometheus_static_targets(self):
        """Verify updating prometheus static targets."""
        config = self.def_config
        config.update(
            {
                "scrape-interval": "15",
                "evaluation-interval": "15",
                "static-targets": "foo:1234 , bar:5678 ",
            }
        )
        hookenv.config.return_value = config
        react_prom.update_prometheus_no_targets()
        react_prom.write_prometheus_config_yml()
        react_prom.write_prometheus_config_def()
        react_prom.validate_config.assert_called_once_with()
        yaml_content = yaml.safe_load(open(self.prom_yml))
        exp_dict = {
            "job_name": "static-targets",
            "static_configs": [
                {
                    "targets": ["foo:1234", "bar:5678"],
                    "labels": {"group": "promoagents-static"},
                }
            ],
        }
        self.assertDictEqual(yaml_content["scrape_configs"][1], exp_dict)

    def test_update_prometheus_federate_config(self):
        """Verify updating prometheus federated config."""
        federate_config = {
            "job_name": "test-federation",
            "targets": ["prometheus-2:9090", "prometheus-a-2:9090"],
            "params": {"match[]": ['{job="prometheus"}', '{someothertag="foo"}']},
        }
        config = self.def_config
        config.update(
            {
                "scrape-interval": "15",
                "evaluation-interval": "15",
                "federate-config": yaml.dump([federate_config]),
            }
        )
        hookenv.config.return_value = config
        react_prom.update_prometheus_no_targets()
        react_prom.write_prometheus_config_yml()
        react_prom.write_prometheus_config_def()
        react_prom.validate_config.assert_called_once_with()
        yaml_content = yaml.safe_load(open(self.prom_yml))
        exp_dict = {
            "job_name": "test-federation",
            "honor_labels": True,
            "metrics_path": "/federate",
            "params": {"match[]": ['{job="prometheus"}', '{someothertag="foo"}']},
            "static_configs": [
                {
                    "targets": ["prometheus-2:9090", "prometheus-a-2:9090"],
                    "labels": {"group": "promoagents-federate"},
                }
            ],
        }
        self.assertDictEqual(yaml_content["scrape_configs"][1], exp_dict)

    def test_update_prometheus_no_alertmanager(self):
        """Verify updating prometheus with no alertmanager."""
        react_prom.update_prometheus_no_alertmanager()
        react_prom.write_prometheus_config_yml()
        # Verify no alertmanagers in config
        with open(self.prom_yml) as fh:
            prometheus_config_dict = yaml.safe_load(fh.read())
        self.assertFalse("alerting" in prometheus_config_dict.keys())

    def test_update_prometheus_alertmanager(self):
        """Verify updating promehteus specifying the alertmanager."""
        srv1_name = "alertmanager"
        srv1_hostport1 = ["foohost1", "fooport1"]
        srv1_hostport2 = ["foohost2", "fooport2"]
        srv1_mock = {srv1_name: [srv1_hostport1, srv1_hostport2]}
        req_mock = ReactInterfaceMock(srv1_mock)
        react_prom.update_prometheus_alertmanager(req_mock)
        react_prom.write_prometheus_config_yml()
        with open(self.prom_yml) as fh:
            prometheus_config_dict = yaml.safe_load(fh.read())
        exp_dict = {
            "alertmanagers": [
                {
                    "static_configs": [
                        {"targets": ["foohost1:fooport1", "foohost2:fooport2"]}
                    ]
                }
            ]
        }
        self.assertTrue(prometheus_config_dict["alerting"] == exp_dict)

    @mock.patch("reactive.prometheus.set_state")
    def test_install_packages_conditionally_called(self, mock_set_state):
        """Verify prometheus.do-install is called only on reconfiguration."""
        # No configs changed, expect no prometheus-do-install
        config = self.def_config
        config.set_changed(
            {"snap_channel": False, "prometheus_registration_listen": False}
        )
        hookenv.config.return_value = config
        react_prom.check_reconfig_prometheus()
        self.assertFalse(
            mock.call("prometheus.do-install") in mock_set_state.call_args_list
        )
        # Mock some configs changed, expect prometheus-do-install
        mock_set_state.reset_mock()
        config.set_changed(
            {"snap_channel": True, "prometheus_registration_listen": False}
        )
        react_prom.check_reconfig_prometheus()
        self.assertTrue(
            mock.call("prometheus.do-install") in mock_set_state.call_args_list
        )

    def test_other_configs(self):
        """Verify config-changed."""
        config = self.def_config
        config.set_changed(
            {"prometheus_registration_listen": True, "custom-rules": True}
        )
        fake_ext_url = "http://foo:1234/bar"
        fake_rules = "BLAH\nBLEH!"
        config.update({"external_url": fake_ext_url, "custom-rules": fake_rules})
        hookenv.config.return_value = config
        react_prom.update_prometheus_no_targets()
        react_prom.write_prometheus_config_yml()
        react_prom.write_prometheus_config_def()
        react_prom.configure_rules()
        # Verify etc/default/prometheus has -web.external-url added
        with open(self.prom_def) as fh:
            self.assertRegexpMatches(
                fh.readline(), "ARGS.*-web.external-url {}.*".format(fake_ext_url)
            )
        # Verify custom.rules file created with expected content
        with open(self.prom_custom_rules) as fh:
            self.assertEqual(fh.read(), fake_rules)

        # Verify generic rules are present
        with open(react_prom.PATHS["generic_rules_path"]) as fh:
            self.assertRegexpMatches(fh.read(), "AllTargets")

    @mock.patch("charms.reactive.bus.ExternalHandler")
    @mock.patch("reactive.prometheus.fetch")
    @mock.patch("reactive.prometheus.install_packages")
    def test_started_after_full_run(
        self, mock_install_packages, mock_fetch, mock_ext_handler
    ):
        """Simulate a full run.

        From zero (actually, 'prometheus.installed') to final state,
        verify we end calling service_restarted('prometheus').
        """
        react_prom.data_changed.return_value = True
        host.service_running.return_value = True
        bus.set_state("prometheus.installed")
        bus.dispatch()
        host.service_running.assert_called_with("snap.promreg.promreg")
        host.service_restart.assert_called_with("snap.promreg.promreg")
